# -*- coding: utf-8 -*-

#imports
from __future__ import division
import numpy as np
import copy
import logging

#create logger
logger = logging.getLogger("Creature")
logging.basicConfig(level=logging.WARNING)

visualPoint = False
finx = 800
finy = 0

########### COORDINATEN KLASSE ###########
class Coord(object):
	'''Koordinaten Umrechner'''
	
	width = 1280	#Bildschirmgröße
	height = 720
	
	def __init__(self, width=1280, height=720):
		self.width = width
		self.height = height
		
	def sim2turt(self, x, y):
		return (x-self.width/3,y)
	
	def sim2pyg(self, x, y):
		return (x+100,y+self.height/2)

########### UNSERE HAUPTKLASSE ###########
class Creature(object):
	'''Bauplan unserer Kreaturen'''
	
	species = [0,0,0,0,0]	#Art der Mutationsrate
	mommiesway = []		#Vererbte Liste von Eltern
	milkyway = []		#unsere Wegbeschreibung, aus der die Kreatur ihre Bewegung bezieht
	survivaltimer = 0	#überlebensdauer
	traveled = 0		#zurückgelegter Weg
	viewpoint = 0		#blickrichtung
	bestAcc = [0,0,0,0,0]	#Statistik fürs Rating
	
	xcor = 0 #x-Achsen Koordinate
	ycor = 0 #y-Achsen Koordinate
	
	global finx, finy, visualPoint
	
	dead = False
	success = False
	
	def __init__(self, species=0, oldway=[]):
		'''Kreatur wird konstruiert'''
		if type(species) == type(10):
			if species == 0:
				self.species = [['R','L','S'],[22.5,17.5],[6,4],3,4]
			if species == 1:
				self.species = [['R','L','S','S','S'],[42.5,27.5],[4.5,3.5],4,2]
			if species == 2:
				self.species = [['R','R','R','L','L','L','S'],[17.5,17.5],[10,5],2,6]
			if species == 3:
				self.species = [['R','L'],[10,10],[8,5],1,0]
		else:
			self.species = species
			
		self.mommiesway = oldway
		self.milkyway = []
		self.bestAcc = [0,0,0,0,0]
		
		logging.info("	Konstruktor: Kreatur wurde erstellt")
	
	def milkAppend(self):
		'''milkyway wird um n Blöcke erweitert'''
		mom = self.survivaltimer//self.species[3] #Code-Leserlichkeit
		if mom >= len(self.mommiesway):	#es gibt keinen Mutterweg
			for i in range(self.species[3]):
				block = []	#Bewegungsblock
				
				#Richtung
				block.append(np.random.choice(self.species[0]))
					
				#Drehung	
				block.append(np.random.randint(self.species[1][0]-self.species[1][1],self.species[1][0]+self.species[1][1]+1))
				
				#Schrittlänge
				block.append(np.random.randint(self.species[2][0]-self.species[2][1],self.species[2][0]+self.species[2][1]+1))
				
				#Aneinanderkettung der Bewegungsblöcke
				self.milkyway.append(block)
		
		################################
		else:	#es gibt einen Mutterweg
			#Richtung
			if self.mommiesway[mom][0]>0:
				direction = [['R'],self.mommiesway[mom][0]]
			elif self.mommiesway[mom][0]<0:
				direction = [['L'],np.absolute(self.mommiesway[mom][0])]
			else:
				direction = [['S'],0]
			
			for i in range(self.species[3]):
				block = []	#Bewegungsblock
				
				#Richtung
				block.append(np.random.choice(direction[0]))
					
				#Drehung
				if direction[1] > self.species[1][0]+self.species[1][1]-self.species[4]:
					direction[1] = self.species[1][0]+self.species[1][1]-self.species[4]
				block.append(np.random.randint(direction[1]-self.species[4],direction[1]+self.species[4]+1))
					
				#Schrittlänge
				if self.mommiesway[mom][1] > self.species[2][0]+self.species[2][1]-(self.species[4]/4):
					self.mommiesway[mom][1] = self.species[2][0]+self.species[2][1]-(self.species[4]/4)
				elif self.mommiesway[mom][1] < self.species[2][0]-self.species[2][1]+(self.species[4]/4):
					self.mommiesway[mom][1] = self.species[2][0]-self.species[2][1]+(self.species[4]/4)
				block.append(np.random.randint(self.mommiesway[mom][1]-(self.species[4]/4),self.mommiesway[mom][1]+(self.species[4]/4)+1))
				
				#Aneinanderkettung der Bewegungsblöcke
				self.milkyway.append(block)
	
	def reset(self):
		'''Setzt die Kreatur zurueck auf Null'''
		self.xcor = -200.0
		self.ycor = 0.0
		self.viewpoint = 0.0
		self.survivaltimer = 0
		self.bestAcc = [0,0,0,0,0]
		self.dead = False
		self.visualPoint = False
		self.milkyway = []
		self.milkAppend()
				
		logging.info("	Kreatur ist auf Null")
	
	def moveFurther(self, moveBlock):
		'''Bewegt die Kreatur einmal'''

		if self.dead:
			logging.debug("	Einheit tot, Bewegung gestoppt!")
			return
		
		if len(moveBlock)>3:
			logging.warning("	Falsches moveBlock Format! (zu lang)")
		elif len(moveBlock)!=3:
			logging.error("	Falsches moveBlock Format! (zu kurz)")
		
		if type(moveBlock[2]) != float and type(moveBlock[2]) != int:
			logging.error("	moveBlock[2]: falscher Typ!")
			return
			
		if type(moveBlock[1]) != float and type(moveBlock[1]) != int:
			logging.error("	moveBlock[1]: falscher Typ!")
			return
		
		if moveBlock[0] != 'R' and moveBlock[0] != 'L' and moveBlock[0] != 'S':
			logging.warning("	moveBlock[0]: keine gültige Richtung! | Kreatur geht geradeaus!")
		
		if moveBlock[0]=='R':	#rechtsdrehung
			self.viewpoint -= moveBlock[1]
		elif moveBlock[0]=='L':	#linksdrehung
			self.viewpoint += moveBlock[1]
		else:					#geradeaus
			pass
		
		#auf dem kreis bleiben:
		if self.viewpoint > 359:
			self.viewpoint -= 360
		if self.viewpoint < 0:
			self.viewpoint += 360
		
		#eigentliche Bewegung im Koordinatensystem
		
		self.ycor += moveBlock[2]*np.sin(self.viewpoint/180*np.pi)
		self.xcor += moveBlock[2]*np.cos(self.viewpoint/180*np.pi)
		
		
		self.survivaltimer += 1
		self.traveled += moveBlock[2]
		
		self.deadly()
		
		self.rating()
			
	def movement(self):
		'''Bewegt die Kreatur mehrere Male'''
		apCounter = self.species[3]
		self.milkAppend()
		for i in self.milkyway:
			self.moveFurther(i)
			if apCounter == 1 and self.survivaltimer < 12 and not self.dead:
				self.milkAppend()
				apCounter = self.species[3]
			else:
				apCounter -= 1
			
			logging.debug("	Bewegungszwischenstand: %s, %s" %(apCounter,self.survivaltimer))
		
		self.dead = True
		self.rating()
		logging.info("	Kreatur ist erschöpft")
	
	def distRating(self):		
		#Distanzrating
		#guckt auf den Abstand zum Ziel, im Vergleich zum Startabstand
		return (100 - (((finx - self.xcor)**2 + (finy - self.ycor)**2)**0.5 * 100) / ((0 - finx)**2 + finy**2)**0.5)
		
	def accuraRating(self):
		#Idealweg, an dem sich ungefähr gerichtet werden soll, um das Ziel zu finden
		bestway = [[0,0],[50,0],[100,0],[150,0],[200,0],[250,0],[300,0],[350,0],[400,0],[450,0],[500,0],[550,0],[600,0],[650,0],[700,0],[750,0],[800,0]]
		
		#Zielstrebigkeit
		#Ermittelung des geringsten Abstandes der Kreatur zum Idealweg
		dista = [ np.linalg.norm(np.array(v) - np.array([self.xcor, self.ycor])) for v in bestway]
		i = np.argmin(np.array(dista))	
		return [(100 - (dista[i]/2)),i]

	def sTimeRating(self):
		#Überlebenszeit
		return ((100 - self.survivaltimer)*2)
		
	def rating(self):
		'''allgemeines Rating'''
		ratDist = 0
		ratSurv = 0
		ratAccur = 0
		
		global visualPoint
		#Norm fürs Rating: 100 - beste; 0 - tolerant; negativ - schlecht
		
		if not self.dead:	#dieses Rating wird nach jeder Bewegung erstellt
			testAcc = self.accuraRating()
			if self.bestAcc[1] < testAcc[1] and self.bestAcc[3] < self.distRating() and (self.bestAcc[0]-testAcc[0])<50:
				self.bestAcc[0] = testAcc[0]
				self.bestAcc[1] = testAcc[1]
				self.bestAcc[2] = self.survivaltimer
				self.bestAcc[3] = self.distRating()
				self.bestAcc[4] += self.bestAcc[1]*1000/self.traveled
				visualPoint = True #visualisierung
			elif self.bestAcc[1] == testAcc[1] and self.bestAcc[0] < testAcc[0] and finAngle(self.xcor,self.ycor,finx,finy,self.viewpoint) == True:
				self.bestAcc[0] = testAcc[0]
				self.bestAcc[2] = self.survivaltimer
				self.bestAcc[4] += 2
				visualPoint = True #visualisierung
			else:
				visualPoint = False #visualisierung
		
		if self.dead:		#dieses Rating wird nur nach dem Tod erstellt
			ratDist = self.distRating()
			logging.debug("	Distanz-Rating ergab: %.2f" %(ratDist))
			
			ratSurv = self.sTimeRating()
			logging.debug("	Überlebenszeit-Rating: %s" %(ratSurv))
			
			ratAccur = self.accuraRating()	#Achtung, hier wird eine [(),()]Liste ausgegeben
			logging.debug("	Zielstrebigkeit: %.2f im Punkt %s" %(ratAccur[0],ratAccur[1]))
			logging.debug("	Bester zielstrebige Wert: %.2f im Punkt %s nach %s Schritten" %(self.bestAcc[0],self.bestAcc[1],self.bestAcc[2]))	#HIER BEDEUTUNG VON bestAcc[] ABLESEN
			logging.info("	Generelle Verbesserungsrating: %i" %(self.bestAcc[4]))
			logging.info("	Todes-Rating wurde erstellt")
			
			return self.bestAcc[4]
	
	def deadly(self):
		'''Tod der Kreatur'''
		if self.xcor > finx-10 and self.xcor < finx+10 and self.ycor > finy-10 and self.ycor < finy+10:
			logging.warning("	Einheit hat Ziel erreicht!")
			self.dead = True
			self.success = True
	
	def legacy(self):
		'''mommiesway und species wird für die nächste Generation vorbereitet'''
		self.mommiesway = []	#überschrieben
		partcount = 0	#wievielte Blockgruppe in milkyway
		
		while partcount*self.species[3] < self.bestAcc[2]:	#milkyway begrenzen?
			summ = [0,0]	#Summe für die Durchschnittsbildung
			block = []	#mommyway-Block
			p = partcount*self.species[3]	#Code-Leserlichkeit
			
			#Durchschnitts-Ermittlung
			for i in range(self.species[3]):
				if self.milkyway[p+i][0] == 'R':
					summ[0] += self.milkyway[p+i][1]
				elif self.milkyway[p+i][0] == 'L':
					summ[0] -= self.milkyway[p+i][1]
				else:
					pass
				
				summ[1] += self.milkyway[p+i][2]
			
			#neuschreiben von mommiesway für Tochter
			block.append(round(summ[0]/self.species[3]))
			block.append(round(summ[1]/self.species[3]))
			self.mommiesway.append(block)
			
			partcount += 1
		

########### WEITERE FUNKTIONEN ###########
def getKids(cMother):
	'''erstellt die neue Generation'''
	cMother.legacy()
	mutateSpec = copy.deepcopy(cMother.species)
	
	if mutateSpec[1][0] > 5 and mutateSpec[1][0] < 85:
		mutateSpec[1][0] = np.random.uniform(-1,1)+mutateSpec[1][0]
	elif mutateSpec[1][0] < 5:
		mutateSpec[1][0] = np.random.random()+mutateSpec[1][0]
	else:
		mutateSpec[1][0] = -np.random.random()+mutateSpec[1][0]
	
	if mutateSpec[1][1] <= mutateSpec[1][0]-1 and mutateSpec[1][1] > 1:
		mutateSpec[1][1] = np.random.uniform(-1,1)+mutateSpec[1][1]
	elif mutateSpec[1][1] < 1:
		mutateSpec[1][1] = np.random.random()+mutateSpec[1][1]
	else:
		mutateSpec[1][1] = -np.random.random()+mutateSpec[1][1]
	
	if mutateSpec[2][0] > 1 and mutateSpec[2][0] < 30:
		mutateSpec[2][0] = np.random.uniform(-1,1)+mutateSpec[2][0]
	elif mutateSpec[2][0] < 1:
		mutateSpec[2][0] = np.random.random()+mutateSpec[2][0]
	else:
		mutateSpec[2][0] = -np.random.random()+mutateSpec[2][0]
	
	if mutateSpec[2][1] <= mutateSpec[2][0]-1 and mutateSpec[2][1] > 1:
		mutateSpec[2][1] = np.random.uniform(-1,1)+mutateSpec[2][1]
	elif mutateSpec[2][1] < 1:
		mutateSpec[2][1] = np.random.random()+mutateSpec[2][1]
	else:
		mutateSpec[2][1] = -np.random.random()+mutateSpec[2][1]
	
	if mutateSpec[4] > 0.5:
		mutateSpec[4] = np.random.uniform(-1,1)/4+mutateSpec[4]
	else:
		mutateSpec[4] = np.random.random()/4+mutateSpec[4]
	
	return Creature(species=mutateSpec, oldway=cMother.mommiesway)

def finAngle(x,y,finx,finy,vp):
	'''berechnet den Winkel von viewpoint zum Ziel'''
	alpha = vp - np.degrees(np.arctan2(finy-y,finx-x))
	if alpha > 270 or alpha < 90:
		return True
	else:
		return False
	

if __name__ == "__main__":
	c1 = Creature(species=3)
	c1.movement()
	print ""
	print c1.milkyway
	print ""
	
	c2 = getKids(c1)
	print c1.mommiesway
	print ""
	print ""

	c2.movement()
	print ""
	print c2.milkyway
	
	c2.legacy()
	print ""
	print c2.mommiesway
